import os

from google.cloud.sql.connector import Connector, IPTypes
import pg8000

import sqlalchemy


def connect_with_connector() -> sqlalchemy.engine.base.Engine:
    """
    Initializes a connection pool for a Cloud SQL instance of Postgres.

    Uses the Cloud SQL Python Connector package.
    """
    # Note: Saving credentials in environment variables is convenient, but not
    # secure - consider a more secure solution such as
    # Cloud Secret Manager (https://cloud.google.com/secret-manager) to help
    # keep secrets safe.

    instance_connection_name = os.environ["INSTANCE_CONNECTION_NAME"]  # e.g. 'project:region:instance'
    db_user = os.environ["DB_USER"]  # e.g. 'my-db-user'
    db_pass = os.environ["DB_PASS"]  # e.g. 'my-db-password'
    db_name = os.environ["DB_NAME"]  # e.g. 'my-database'
    print(instance_connection_name, db_user, db_pass[-5::], db_name)

    ip_type = IPTypes.PRIVATE if os.environ.get("PRIVATE_IP") else IPTypes.PUBLIC

    # initialize Cloud SQL Python Connector object
    connector = Connector()

    def getconn() -> pg8000.dbapi.Connection:
        conn: pg8000.dbapi.Connection = connector.connect(
            instance_connection_name,
            "pg8000",
            user=db_user,
            password=db_pass,
            db=db_name,
            ip_type=ip_type,
        )
        return conn

    # The Cloud SQL Python Connector can be used with SQLAlchemy
    # using the 'creator' argument to 'create_engine'
    pool = sqlalchemy.create_engine(
        "postgresql+pg8000://",
        creator=getconn,
        # [START_EXCLUDE]
        # Pool size is the maximum number of permanent connections to keep.
        pool_size=5,

        # Temporarily exceeds the set pool_size if no connections are available.
        max_overflow=2,

        # The total number of concurrent connections for your application will be
        # a total of pool_size and max_overflow.

        # 'pool_timeout' is the maximum number of seconds to wait when retrieving a
        # new connection from the pool. After the specified amount of time, an
        # exception will be thrown.
        pool_timeout=30,  # 30 seconds

        # 'pool_recycle' is the maximum number of seconds a connection can persist.
        # Connections that live longer than the specified amount of time will be
        # re-established
        pool_recycle=1800,  # 30 minutes
        # [END_EXCLUDE]
    )
    return pool



from gandai import query
from gandai.models import Search
import requests
import os
from dotenv import load_dotenv

load_dotenv()


class GrataWrapper:
    HEADERS = {
        "Authorization": os.getenv("GRATA_API_TOKEN"),
        "Content-Type": "application/json",
    }

    def __name__(self):
        return "GrataWrapper"

    def find_similar(domain: str, search: Search) -> list:
        api_filters = GrataWrapper._get_api_filter(search)
        response = requests.post(
            "https://search.grata.com/api/v1/search-similar/",
            headers=GrataWrapper.HEADERS,
            json={
                "domain": domain,
                "employees_range": api_filters["employees_range"],
                "headquarters": api_filters["headquarters"],
            },
        )
        data = response.json()
        print("find_similar:", data)
        data["companies"] = data.get("results",[])  # asking grata about this
        
        return data['companies']

    def find_by_criteria(search: Search) -> dict:
        api_filters = GrataWrapper._get_api_filter(search)
        response = requests.post(
            "https://search.grata.com/api/v1/search/",
            headers=GrataWrapper.HEADERS,
            json={
                "op": "any",
                "include": search.inclusion["keywords"],
                "exclude": search.exclusion["keywords"],
                "employees_range": api_filters["employees_range"],
                "ownership": api_filters["ownership"],
                "headquarters": api_filters["headquarters"],
            },
        )
        data = response.json()
        print("find_by_criteria: ", data)
        return data['companies']

    def enrich(domain: str) -> dict:
        response = requests.post(
            "https://search.grata.com/api/v1/enrich/",
            headers=GrataWrapper.HEADERS,
            json={"domain": domain},
        )
        data = response.json()
        data['linkedin'] = data.get('social_linkedin')
        return data

    def _get_api_filter(search: Search) -> dict:
        STATES = {
            "AL": "Alabama",
            "AK": "Alaska",
            "AZ": "Arizona",
            "AR": "Arkansas",
            "CA": "California",
            "CO": "Colorado",
            "CT": "Connecticut",
            "DE": "Delaware",
            "FL": "Florida",
            "GA": "Georgia",
            "HI": "Hawaii",
            "ID": "Idaho",
            "IL": "Illinois",
            "IN": "Indiana",
            "IA": "Iowa",
            "KS": "Kansas",
            "KY": "Kentucky",
            "LA": "Louisiana",
            "ME": "Maine",
            "MD": "Maryland",
            "MA": "Massachusetts",
            "MI": "Michigan",
            "MN": "Minnesota",
            "MS": "Mississippi",
            "MO": "Missouri",
            "MT": "Montana",
            "NE": "Nebraska",
            "NV": "Nevada",
            "NH": "New Hampshire",
            "NJ": "New Jersey",
            "NM": "New Mexico",
            "NY": "New York",
            "NC": "North Carolina",
            "ND": "North Dakota",
            "OH": "Ohio",
            "OK": "Oklahoma",
            "OR": "Oregon",
            "PA": "Pennsylvania",
            "RI": "Rhode Island",
            "SC": "South Carolina",
            "SD": "South Dakota",
            "TN": "Tennessee",
            "TX": "Texas",
            "UT": "Utah",
            "VT": "Vermont",
            "VA": "Virginia",
            "WA": "Washington",
            "WV": "West Virginia",
            "WI": "Wisconsin",
            "WY": "Wyoming",
        }

        COUNTRIES = {
            "USA": "United States",
            "CAN": "Canada",
            "MEX": "Mexico",
            "GBR": "United Kingdom",
        }

        def _ownership_filter() -> dict:
            # API wants inclusion, but product wants reverse
            OWNERSHIP = [
                "bootstrapped",
                "investor_backed",
                "public",
                "public_subsidiary",
                "private_subsidiary",
                "private_equity",
                "private_equity_add_on",
            ]
            return list(set(OWNERSHIP) - set(search.exclusion.get("ownership",[])))

        def _hq_include() -> list:
            include = []
            for state in search.inclusion.get("state", []):
                # NB: API wants full state name, but product wants code
                include.append({"state": STATES[state]})
            for country in search.inclusion.get("country", []):
                include.append({"country": COUNTRIES[country]})
            return include

        def _hq_exclude() -> list:
            exclude = []
            for state in search.exclusion.get("state", []):
                exclude.append({"state": state})
            return exclude

        return {
            "op": "any",
            "include": search.inclusion.get("keywords", []),
            "exclude": search.exclusion.get("keywords", []),
            "employees_range": search.inclusion.get("employees_range", []),
            "ownership": _ownership_filter(),
            "headquarters": {
                "include": _hq_include(),
                "exclude": _hq_exclude(),
            },
        }


class SourceScrubWrapper:
    def find_similar(domain: str, search: Search) -> dict:
        pass

    def find_by_criteria(search: Search) -> dict:
        pass

    def enrich(domain: str) -> dict:
        pass



import pandas as pd
from dacite import from_dict
from dataclasses import dataclass, field
from concurrent.futures import ThreadPoolExecutor

from gandai.models import Event, Company, Checkpoint
from gandai import query

from gandai.sources import GrataWrapper as grata


def process_event(event_id: int) -> None:
    """
    May trigger additional targets adding to inbox, or something else
    (e.g. a notification)
    """

    def _insert_companies(companies: list, existing_domains: list) -> None:
        # rewrite this in less db txns
        for company in companies:
            if company.get("domain") is None:
                print(f"Missing domain: {company}. Skipping")
                continue

            if company["domain"] in existing_domains:
                print(f"Skipping {company['domain']} as already a target")
                continue
            print(f"Adding {company['domain']} as target")
            query.insert_company(
                Company(
                    domain=company["domain"],
                    name=company.get("name"),
                    description=company.get("description"),
                )
            )
            query.insert_event(
                Event(
                    search_uid=search_uid,
                    domain=company.get("domain"),
                    actor_key="grata",
                    type="create",
                )
            )

    # sets up better to do this threaded
    e = query.find_event_by_id(event_id)
    search_uid = e.search_uid
    if e.type == "create":
        pass
    elif e.type == "advance":
        # enrich the company
        company = query.find_company_by_domain(e.domain)
        # might consider a check here to see if the company is already enriched
        resp = grata.enrich(company.domain)
        if resp.get("status") == 404:
            print(f"{company} not found")
        else:
            print(resp)
            company.name = resp.get("name")
            company.description = resp.get("description")
            company.meta = {**company.meta, **resp}  # merge 3.5+
            query.update_company(company)

    elif e.type == "validate":
        search = query.find_search_by_uid(search_uid)
        _insert_companies(
            companies=grata.find_similar(domain=e.domain, search=search),
            existing_domains=query.target(search_uid=search_uid)["domain"].tolist(),
        )
    elif e.type == "send":
        pass
    elif e.type == "accept":
        pass
    elif e.type == "reject":
        pass
    elif e.type == "conflict":
        pass
    elif e.type == "criteria":
        print("criteria search here we gooo")
        search = query.find_search_by_uid(search_uid)
        _insert_companies(
            companies=grata.find_by_criteria(search),
            existing_domains=query.target(search_uid=search_uid)["domain"].tolist(),
        )

    # finally, record we processed the event
    query.insert_checkpoint(Checkpoint(event_id=e.id))
    print(f"processed: {e}")


def process_events(search_uid: int) -> int:
    """
    Process all events for a given search
    """

    events = query.event(search_uid=search_uid)
    checkpoints = query.checkpoint(search_uid=search_uid)

    q = list(set(events["id"].tolist()) - set(checkpoints["event_id"].tolist()))
    
    with ThreadPoolExecutor(max_workers=4) as executor:
        executor.map(process_event, q)
    
    return len(q)



import json
import pandas as pd
from dataclasses import asdict
from dacite import from_dict
from gandai.models import (
    Event,
    Company,
    EventType,
    Actor,
    Search,
    Checkpoint,
)


import sqlalchemy

from google.cloud.sql.connector import Connector
from dotenv import load_dotenv

load_dotenv()

from gandai.db import connect_with_connector

db = connect_with_connector()


### WRITES ###


def insert_company(company: Company):
    with db.connect() as con:
        statement = sqlalchemy.text(
            """
                INSERT INTO company (domain, name, description) 
                VALUES(:domain, :name, :description)
                ON CONFLICT DO NOTHING
            """
        )
        con.execute(statement, asdict(company))
        con.commit()
    return company


def insert_event(event: Event) -> Event:
    with db.connect() as con:
        statement = sqlalchemy.text(
            """
                INSERT INTO event (search_uid, domain, actor_key, type, data) 
                VALUES(:search_uid, :domain, :actor_key, :type, :data)
                ON CONFLICT DO NOTHING
                RETURNING id
            """
        )
        obj = asdict(event)
        obj["data"] = json.dumps(obj["data"])
        result = con.execute(statement, obj)
        # print(result.first())
        _id = result.first()
        event.id = _id[0] if _id else None
        con.execute(sqlalchemy.text("REFRESH MATERIALIZED VIEW target"))
        con.commit()
    return event


def insert_actor(actor: Actor) -> Actor:
    with db.connect() as con:
        statement = sqlalchemy.text(
            """
                INSERT INTO actor (key, type, name) 
                VALUES(:key, :type, :name)
                ON CONFLICT DO NOTHING
            """
        )
        obj = asdict(actor)
        con.execute(statement, obj)
        con.commit()
    return actor


def insert_search(search: Search) -> Search:
    with db.connect() as con:
        statement = sqlalchemy.text(
            """
                INSERT INTO search (uid, client_domain, label, meta, inclusion, exclusion, sort) 
                VALUES(:uid, :client_domain, :label, :meta, :inclusion, :exclusion, :sort)
                ON CONFLICT DO NOTHING
            """
        )
        obj = asdict(search)
        obj["meta"] = json.dumps(obj["meta"])
        obj["inclusion"] = json.dumps(obj["inclusion"])
        obj["exclusion"] = json.dumps(obj["exclusion"])
        obj["sort"] = json.dumps(obj["sort"])
        con.execute(statement, obj)
        con.commit()
    return search


def insert_checkpoint(checkpoint: Checkpoint) -> Checkpoint:
    with db.connect() as con:
        statement = sqlalchemy.text(
            """
                INSERT INTO checkpoint (event_id) 
                VALUES(:event_id)
            """
        )
        con.execute(statement, asdict(checkpoint))
        con.commit()
    return checkpoint


### READS ###


def search():
    with db.connect() as conn:
        result = conn.execute(
            sqlalchemy.text("SELECT *, meta->>'group' as group FROM search")
        )
        df = pd.DataFrame(result.fetchall(), columns=result.keys())
    return df


def target(search_uid: int, last_event_type: str = None):
    with db.connect() as conn:
        if last_event_type is not None:
            statement = "SELECT * FROM target WHERE search_uid = :search_uid AND last_event_type = :last_event_type"
            result = conn.execute(
                sqlalchemy.text(statement),
                {"search_uid": search_uid, "last_event_type": last_event_type},
            )

        else:
            statement = "SELECT * FROM target WHERE search_uid = :search_uid"
            result = conn.execute(
                sqlalchemy.text(statement),
                {"search_uid": search_uid},
            )

        targets = pd.DataFrame(result.fetchall(), columns=result.keys())

    comments = comment_by_domain(search_uid)
    targets = targets.merge(comments, on="domain", how="left")

    # handle sorting
    search = find_search_by_uid(search_uid)
    targets = targets.sort_values(
        by=search.sort.get("field", "domain"),
        ascending=search.sort.get("order") == "asc",
    )

    return targets


def target_count(search_uid: int) -> pd.DataFrame:
    with db.connect() as conn:
        statement = """
                SELECT last_event_type, count(*)
                FROM target
                WHERE search_uid = :search_uid
                GROUP BY last_event_type
            """
        result = conn.execute(
            sqlalchemy.text(statement),
            {"search_uid": search_uid},
        )
        df = pd.DataFrame(result.fetchall(), columns=result.keys())
    return df


def event(search_uid: int) -> pd.DataFrame:
    with db.connect() as conn:
        statement = """
                SELECT *
                FROM event
                WHERE search_uid = :search_uid
            """
        result = conn.execute(sqlalchemy.text(statement), {"search_uid": search_uid})
        df = pd.DataFrame(result.fetchall(), columns=result.keys())
    return df


def company() -> pd.DataFrame:
    with db.connect() as conn:
        statement = """
                SELECT *
                FROM company
            """
        result = conn.execute(sqlalchemy.text(statement))
        df = pd.DataFrame(result.fetchall(), columns=result.keys())
    return df


def checkpoint(search_uid: int) -> pd.DataFrame:
    with db.connect() as conn:
        statement = """
                SELECT *
                FROM checkpoint
                JOIN event ON checkpoint.event_id = event.id
                WHERE search_uid = :search_uid
            """
        result = conn.execute(sqlalchemy.text(statement), {"search_uid": search_uid})
        df = pd.DataFrame(result.fetchall(), columns=result.keys())
    return df


def comment_by_domain(search_uid: int) -> pd.DataFrame:
    with db.connect() as conn:
        statement = """
                SELECT *, data->>'comment' AS comment
                FROM event e
                WHERE 
                    search_uid = :search_uid AND
                    type = 'comment'
            """
        result = conn.execute(
            sqlalchemy.text(statement),
            {"search_uid": search_uid},
        )
        df = pd.DataFrame(result.fetchall(), columns=result.keys())

    return df.groupby("domain").agg({"comment": lambda x: list(x)}).reset_index()


### FINDERS -> dataclass ###


def find_search_by_uid(search_uid: int) -> Search:
    with db.connect() as conn:
        statement = """
                SELECT *
                FROM search
                WHERE uid = :search_uid
            """
        result = conn.execute(sqlalchemy.text(statement), {"search_uid": search_uid})

    if result.rowcount == 0:
        return None
    else:
        obj = dict(zip(result.keys(), result.fetchone()))
        return from_dict(Search, obj)


def find_company_by_domain(domain: str) -> Company:
    with db.connect() as conn:
        statement = """
                SELECT *
                FROM company
                WHERE domain = :domain
            """
        result = conn.execute(sqlalchemy.text(statement), {"domain": domain})
        # obj = dict(zip(result.keys(), result.fetchone()))
    if result.rowcount == 0:
        return None
    else:
        obj = dict(zip(result.keys(), result.fetchone()))
        return from_dict(Company, obj)


def find_event_by_id(event_id: int) -> Event:
    with db.connect() as conn:
        statement = """
                SELECT *
                FROM event
                WHERE id = :event_id
            """
        result = conn.execute(sqlalchemy.text(statement), {"event_id": event_id})
        # obj = dict(zip(result.keys(), result.fetchone()))
    if result.rowcount == 0:
        return None
    else:
        obj = dict(zip(result.keys(), result.fetchone()))
        return from_dict(Event, obj)

### UPDATE ###


def update_company(company: Company) -> None:
    with db.connect() as conn:
        statement = """
            UPDATE company
            SET
                name = :name,
                description = :description,
                meta = :meta
            WHERE domain = :domain
            """

        conn.execute(
            sqlalchemy.text(statement),
            {
                "name": company.name,
                "description": company.description,
                "domain": company.domain,
                "meta": json.dumps(company.meta),
            },
        )
        conn.execute(sqlalchemy.text("REFRESH MATERIALIZED VIEW target"))
        conn.commit()


def update_search(search: Search) -> None:
    with db.connect() as conn:
        conn.execute(
            sqlalchemy.text(
                """
                UPDATE search
                SET
                    sort = :sort,
                    inclusion = :inclusion,
                    exclusion = :exclusion
                WHERE uid = :uid
                """
            ),
            {
                "sort": json.dumps(search.sort),
                "inclusion": json.dumps(search.inclusion),
                "exclusion": json.dumps(search.exclusion),
                "uid": search.uid,
            },
        )
        conn.commit()



from gandai import query
from gandai.models import Company, Actor, Search, Event
from gandai.migrations.dealcloud import migrate_engagements

def seed_actors():
    actors = {
        "7138248581": "Parker",
        "6508620943": "Gabe",
        "9413500954": "Skye",
        "3102835279": "Chris",
        "5126571681": "Brandon",
        "5125659474": "Jack",
        "3125740050": "Jim",
        "4805706789": "Eli",
        "grata": "Grata Bot",
        "dealcloud": "DealCloud Bot",
        "chatgpt": "ChatGPT",
    }

    for key, name in actors.items():
        actor = query.insert_actor(Actor(key=key, type="research", name=name))
        print(f"Added: {actor}")

def seed_hello_world():
    search = query.insert_search(
        Search(
            uid=1,
            client_domain="parkerholcomb.com",
            label="Hello World",
            meta={"research": "Parker", "status": "In Progress"},
            inclusion={
                "keywords": [],
                "employees_range": [0, 100],
                "country": ["USA"],
            },
            exclusion={"keywords": [], "state": []},
            sort={"field": "domain", "order": "desc"},
        )
    )

    company = query.insert_company(
        Company(
            domain="grata.com",
            name="Grata",
            description="Grata is a thing.",
        )
    )

    e = query.insert_event(
        Event(
            search_uid=search.uid,
            domain=company.domain,
            actor_key="7138248581",
            type="create",
        )
    )


def main():
    seed_actors()
    seed_hello_world()
    

if __name__ == "__main__":
    main()
